"""
Gini based Inequality Metrics
"""

__author__ = "Sergio J. Rey <srey@asu.edu> "

import numpy
from scipy.stats import norm

__all__ = ["Gini", "Gini_Spatial"]


def _gini(x):
    """
    Memory efficient calculation of Gini coefficient
    in relative mean difference form.

    Parameters
    ----------

    x : array-like

    Attributes
    ----------

    g : float
        Gini coefficient.

    Notes
    -----
    Based on
    http://www.statsdirect.com/help/default.htm#nonparametric_methods/gini.htm.

    """
    n = len(x)
    try:
        x_sum = x.sum()
    except AttributeError:
        x = numpy.asarray(x)
        x_sum = x.sum()
    n_x_sum = n * x_sum
    r_x = (2.0 * numpy.arange(1, len(x) + 1) * x[numpy.argsort(x)]).sum()
    return (r_x - n_x_sum - x_sum) / n_x_sum


class Gini:
    """
    Classic Gini coefficient in absolute deviation form.

    Parameters
    ----------

    y : numpy.array
        An array in the shape :math:`(n,1)` containing the attribute values.

    Attributes
    ----------

    g : float
       Gini coefficient.

    """

    def __init__(self, x):
        self.g = _gini(x)


class Gini_Spatial:  # noqa N801
    """
    Spatial Gini coefficient.

    Provides for computationally based inference regarding the contribution of
    spatial neighbor pairs to overall inequality across a set of regions.
    See :cite:`Rey_2013_sea`.

    Parameters
    ----------

    y : numpy.array
        An array in the shape :math:`(n,1)` containing the attribute values.
    w : libpysal.weights.W
        Binary spatial weights object.
    permutations : int (default 99)
       The number of permutations for inference.

    Attributes
    ----------

    g : float
       Gini coefficient.
    wg : float
       Neighbor inequality component (geographic inequality).
    wcg : float
       Non-neighbor inequality component (geographic complement inequality).
    wcg_share : float
       Share of inequality in non-neighbor component.
    p_sim : float
       (If ``permuations > 0``) pseudo :math:`p`-value for spatial gini.
    e_wcg : float
       (If ``permuations > 0``) expected value of non-neighbor
       inequality component (level) from permutations.
    s_wcg : float
        (If ``permuations > 0``) standard deviation non-neighbor
        inequality component (level) from permutations.
    z_wcg : float
        (If ``permuations > 0``) z-value non-neighbor inequality
        component (level) from permutations.
    p_z_sim : float
        (If ``permuations > 0``) pseudo :math:`p`-value based on
        standard normal approximation of permutation based values.

    Examples
    --------

    >>> import libpysal
    >>> import numpy
    >>> from inequality.gini import Gini_Spatial

    Use data from the 32 Mexican States, decade frequency 1940-2010.

    >>> f = libpysal.io.open(libpysal.examples.get_path('mexico.csv'))
    >>> vnames = [f'pcgdp{dec}' for dec in range(1940, 2010, 10)]
    >>> y = numpy.transpose(numpy.array([f.by_col[v] for v in vnames]))

    Define regime neighbors.

    >>> regimes = numpy.array(f.by_col('hanson98'))
    >>> w = libpysal.weights.block_weights(regimes, silence_warnings=True)
    >>> numpy.random.seed(12345)
    >>> gs = Gini_Spatial(y[:,0], w)

    >>> gs.p_sim
    0.04

    >>> gs.wcg
    4353856.0

    >>> gs.e_wcg
    4170356.7474747472

    Thus, the amount of inequality between pairs of states that are not in the
    same regime (neighbors) is significantly higher than what is expected
    under the null of random spatial inequality.

    """

    def __init__(self, x, w, permutations=99):
        x = numpy.asarray(x)
        g = _gini(x)
        self.g = g
        n = len(x)
        den = x.mean() * 2 * n**2
        d = g * den
        wg = self._calc(x, w)
        wcg = d - wg
        self.g = g
        self.wcg = wcg
        self.wg = wg
        self.dtotal = d
        self.den = den
        self.wcg_share = wcg / den

        if permutations:
            ids = numpy.arange(n)
            wcgp = numpy.zeros((permutations,))
            for perm in range(permutations):
                numpy.random.shuffle(ids)
                wcgp[perm] = d - self._calc(x[ids], w)
            above = wcgp >= self.wcg
            larger = above.sum()
            if (permutations - larger) < larger:
                larger = permutations - larger
            self.wcgp = wcgp
            self.p_sim = (larger + 1.0) / (permutations + 1.0)
            self.e_wcg = wcgp.mean()
            self.s_wcg = wcgp.std()
            self.z_wcg = (self.wcg - self.e_wcg) / self.s_wcg
            self.p_z_sim = 1.0 - norm.cdf(self.z_wcg)

    def _calc(self, x, w):
        sad_sum = 0.0
        for i, js in w.neighbors.items():
            sad_sum += numpy.abs(x[i] - x[js]).sum()
        return sad_sum



import numpy
import pytest

from inequality._indices import (
    abundance,
    ellison_glaeser_egg,
    ellison_glaeser_egg_pop,
    fractionalization_gs,
    gini_gi,
    gini_gi_m,
    gini_gig,
    herfindahl_hd,
    hoover_hi,
    isolation_ii,
    isolation_isg,
    margalev_md,
    maurel_sedillot_msg,
    maurel_sedillot_msg_pop,
    menhinick_mi,
    modified_segregation_msg,
    polarization,
    segregation_gsg,
    shannon_se,
    similarity_w_wd,
    simpson_sd,
    simpson_so,
    theil_th,
    theil_th_brute,
)

x = numpy.array([[0, 1, 2], [0, 2, 4], [0, 0, 3]])

numpy.random.seed(0)
y = numpy.random.randint(1, 10, size=(4, 3))

numpy.random.seed(0)
tau = numpy.random.uniform(size=(3, 3))
numpy.fill_diagonal(tau, 0.0)
tau = (tau + tau.T) / 2

numpy.random.seed(0)
z = numpy.random.randint(10, 50, size=(3, 4))

numpy.random.seed(0)
v = numpy.random.uniform(0, 1, size=(4,))


class TestAbundance:
    def test_abundance(self):
        known = 2
        observed = abundance(x)
        assert known == observed


class TestMargalevMD:
    def test_margalev_md(self):
        known = 0.40242960438184466
        observed = margalev_md(x)
        assert known == pytest.approx(observed)


class TestMenhinickMI:
    def test_menhinick_mi(self):
        known = 0.2886751345948129
        observed = menhinick_mi(x)
        assert known == pytest.approx(observed)


class TestSimpsonSO:
    def test_simpson_so(self):
        known = 0.5909090909090909
        observed = simpson_so(x)
        assert known == pytest.approx(observed)


class TestSimpsonSD:
    def test_simpson_sd(self):
        known = 0.40909090909090906
        observed = simpson_sd(x)
        assert known == pytest.approx(observed)


class TestHerfindahlHD:
    def test_herfindahl_hd(self):
        known = 0.625
        observed = herfindahl_hd(x)
        assert known == pytest.approx(observed)


class TestTheilTH:
    def test_theil_th(self):
        known = 0.15106563978903298
        observed = theil_th(x, ridz=True)
        assert known == pytest.approx(observed)

        with pytest.warns(RuntimeWarning, match="divide by zero encountered"):
            observed = theil_th(x, ridz=False)
        assert numpy.isnan(observed)

        # test brute comparison
        known = theil_th_brute(x, ridz=True)
        observed = theil_th(x, ridz=True)
        assert known == pytest.approx(observed)

        with pytest.warns(RuntimeWarning, match="divide by zero encountered"):
            observed = theil_th_brute(x, ridz=False)
        assert numpy.isnan(observed)


class TestFractionalizationGS:
    def test_fractionalization_gs(self):
        known = 0.375
        observed = fractionalization_gs(x)
        assert known == pytest.approx(observed)


class TestPolarization:
    def test_polarization(self):
        with pytest.raises(RuntimeError, match="Not currently implemented."):
            polarization(None)


class TestShannonSE:
    def test_shannon_se(self):
        known = 1.094070862104929
        observed = shannon_se(y)
        assert known == pytest.approx(observed)


class TestGiniGI:
    def test_gini_gi(self):
        known = 0.05128205128205132
        observed = gini_gi(y)
        assert known == pytest.approx(observed)


class TestGiniGIG:
    def test_gini_gig(self):
        known = numpy.array([0.125, 0.32894737, 0.18181818])
        observed = gini_gig(y)
        numpy.testing.assert_array_almost_equal(known, observed)


class TestGiniGIM:
    def test_gini_gi_m(self):
        known = 0.05128205128205132
        observed = gini_gi_m(y)
        assert known == pytest.approx(observed)


class TestHooverHI:
    def test_hoover_hi(self):
        known = 0.041025641025641046
        observed = hoover_hi(y)
        assert known == pytest.approx(observed)


class TestSimilarityWWD:
    def test_similarity_w_wd(self):
        known = 0.5818596340322582
        observed = similarity_w_wd(y, tau)
        assert known == pytest.approx(observed)


class TestSegregationGSG:
    def test_segregation_gsg(self):
        known = numpy.array([0.18292683, 0.24713959, 0.09725159])
        observed = segregation_gsg(y)
        numpy.testing.assert_array_almost_equal(known, observed)


class TestModifiedSegregationMSG:
    def test_modified_segregation_msg(self):
        known = numpy.array([0.0852071, 0.10224852, 0.0435503])
        observed = modified_segregation_msg(y)
        numpy.testing.assert_array_almost_equal(known, observed)


class TestIsolationISG:
    def test_isolation_isg(self):
        known = numpy.array([1.0732699, 1.21995329, 1.0227105])
        observed = isolation_isg(y)
        numpy.testing.assert_array_almost_equal(known, observed)


class TestIsolationII:
    def test_isolation_ii(self):
        known = numpy.array([1.1161596, 1.31080357, 1.03432983])
        observed = isolation_ii(y)
        numpy.testing.assert_array_almost_equal(known, observed)


class TestEllisonGlaeserEGG:
    def test_ellison_glaeser_egg(self):
        known = numpy.array([0.0544994, 0.01624183, 0.01014058, 0.02880251])
        observed = ellison_glaeser_egg(z)
        numpy.testing.assert_array_almost_equal(known, observed)

        known = numpy.array([-1.0617873, -2.39452501, -1.45991648, -1.11740985])
        observed = ellison_glaeser_egg(z, hs=v)
        numpy.testing.assert_array_almost_equal(known, observed)


class TestEllisonGlaeserEGGPop:
    def test_ellison_glaeser_egg_pop(self):
        known = numpy.array([-0.02150826, 0.01329858, -0.03894556])
        observed = ellison_glaeser_egg_pop(y)
        numpy.testing.assert_array_almost_equal(known, observed)


class TestMaurelSedillotMSG:
    def test_maurel_sedillot_msg(self):
        known = numpy.array([0.07858256, 0.03597749, 0.03937436, -0.00904911])
        observed = maurel_sedillot_msg(z)
        numpy.testing.assert_array_almost_equal(known, observed)

        known = numpy.array([-1.01010171, -2.32421555, -1.38868998, -1.20049894])
        observed = maurel_sedillot_msg(z, hs=v.round(3))
        numpy.testing.assert_array_almost_equal(known, observed)


class TestMaurelSedillotMSGPop:
    def test_maurel_sedillot_msg_pop(self):
        known = numpy.array([-0.05503571, 0.04414672, -0.02866628])
        observed = maurel_sedillot_msg_pop(y)
        numpy.testing.assert_array_almost_equal(known, observed)



import libpysal
import numpy

from inequality.theil import Theil, TheilD, TheilDSim


class TestTheil:
    def test___init__(self):
        f = libpysal.io.open(libpysal.examples.get_path("mexico.csv"))
        vnames = [f"pcgdp{dec}" for dec in range(1940, 2010, 10)]
        y = numpy.transpose(numpy.array([f.by_col[v] for v in vnames]))
        theil_y = Theil(y)
        numpy.testing.assert_almost_equal(
            theil_y.T,
            numpy.array(
                [
                    0.20894344,
                    0.15222451,
                    0.10472941,
                    0.10194725,
                    0.09560113,
                    0.10511256,
                    0.10660832,
                ]
            ),
        )


class TestTheilD:
    def test___init__(self):
        f = libpysal.io.open(libpysal.examples.get_path("mexico.csv"))
        vnames = [f"pcgdp{dec}" for dec in range(1940, 2010, 10)]
        y = numpy.transpose(numpy.array([f.by_col[v] for v in vnames]))
        regimes = numpy.array(f.by_col("hanson98"))
        theil_d = TheilD(y, regimes)
        numpy.testing.assert_almost_equal(
            theil_d.bg,
            numpy.array(
                [
                    0.0345889,
                    0.02816853,
                    0.05260921,
                    0.05931219,
                    0.03205257,
                    0.02963731,
                    0.03635872,
                ]
            ),
        )


class TestTheilDSim:
    def test___init__(self):
        f = libpysal.io.open(libpysal.examples.get_path("mexico.csv"))
        vnames = [f"pcgdp{dec}" for dec in range(1940, 2010, 10)]
        y = numpy.transpose(numpy.array([f.by_col[v] for v in vnames]))
        regimes = numpy.array(f.by_col("hanson98"))
        numpy.random.seed(10)
        theil_ds = TheilDSim(y, regimes, 999)
        numpy.testing.assert_almost_equal(
            theil_ds.bg_pvalue,
            numpy.array([0.4, 0.344, 0.001, 0.001, 0.034, 0.072, 0.032]),
        )



"""Theil Inequality metrics

"""

__author__ = "Sergio J. Rey <srey@asu.edu> "

import numpy

__all__ = ["Theil", "TheilD", "TheilDSim"]

SMALL = numpy.finfo("float").tiny


class Theil:
    """
    Classic Theil measure of inequality.

    .. math::

        T = \\sum_{i=1}^n
            \\left( \\frac{y_i}{\\sum_{i=1}^n y_i} \\ln
                \\left[ N \\frac{y_i}{\\sum_{i=1}^n y_i}\\right]
            \\right
        )

    Parameters
    ----------

    y : numpy.array
        An array in the shape :math:`(n,t)` or :math:`(n,)`
        with :math:`n` taken as the observations across which inequality is
        calculated.  If ``y`` is :math:`(n,)` then a scalar inequality value is
        determined. If ``y`` is :math:`(n,t)` then an array of inequality values are
        determined, one value for each column in ``y``.

    Attributes
    ----------

    T : numpy.array
        An array in the shape :math:`(t,)` or :math:`(1,)`
        containing Theil's *T* for each column of ``y``.

    Notes
    -----
    This computation involves natural logs. To prevent ``ln[0]`` from occurring, a
    small value is added to each element of ``y`` before beginning the computation.

    Examples
    --------

    >>> import libpysal
    >>> import numpy
    >>> from inequality.theil import Theil

    >>> f = libpysal.io.open(libpysal.examples.get_path('mexico.csv'))
    >>> vnames = [f'pcgdp{dec}' for dec in range(1940, 2010, 10)]
    >>> y = numpy.array([f.by_col[v] for v in vnames]).T
    >>> theil_y = Theil(y)

    >>> theil_y.T
    array([0.20894344, 0.15222451, 0.10472941, 0.10194725, 0.09560113,
           0.10511256, 0.10660832])

    """

    def __init__(self, y):
        n = len(y)
        y = y + SMALL * (y == 0)  # can't have 0 values
        yt = y.sum(axis=0)
        s = y / (yt * 1.0)
        lns = numpy.log(n * s)
        slns = s * lns
        t = sum(slns)
        self.T = t


class TheilD:
    """Decomposition of Theil's *T* based on partitioning of
    observations into exhaustive and mutually exclusive groups.

    Parameters
    ----------

    y : numpy.array
        An array in the shape :math:`(n,t)` or :math:`(n,)`
        with :math:`n` taken as the observations across which inequality is
        calculated.  If ``y`` is :math:`(n,)` then a scalar inequality value is
        determined. If ``y`` is :math:`(n,t)` then an array of inequality values are
        determined, one value for each column in ``y``.
    partition : numpy.array
        An array in the shape :math:`(n,)` of elements indicating which partition
        each observation belongs to. These are assumed to be exhaustive.

    Attributes
    ----------

    T : numpy.array
        An array in the shape :math:`(t,)` or :math:`(1,)`
        containing the global inequality *T*.
    bg : numpy.array
        An array in the shape :math:`(n,t)` or :math:`(n,)`
        representing between group inequality.
    wg : numpy.array
        An array in the shape :math:`(n,t)` or :math:`(n,)`
        representing within group inequality.

    Examples
    --------

    >>> import libpysal
    >>> import numpy
    >>> from inequality.theil import TheilD

    >>> f = libpysal.io.open(libpysal.examples.get_path('mexico.csv'))
    >>> vnames = [f'pcgdp{dec}' for dec in range(1940, 2010, 10)]
    >>> y = numpy.array([f.by_col[v] for v in vnames]).T
    >>> regimes = numpy.array(f.by_col('hanson98'))
    >>> theil_d = TheilD(y, regimes)

    >>> theil_d.bg
    array([0.0345889 , 0.02816853, 0.05260921, 0.05931219, 0.03205257,
           0.02963731, 0.03635872])

    >>> theil_d.wg
    array([0.17435454, 0.12405598, 0.0521202 , 0.04263506, 0.06354856,
           0.07547525, 0.0702496 ])

    """

    def __init__(self, y, partition):
        groups = numpy.unique(partition)
        T = Theil(y).T  # noqa N806
        ytot = y.sum(axis=0)

        # group totals
        gtot = numpy.array([y[partition == gid].sum(axis=0) for gid in groups])

        if ytot.size == 1:  # y is 1-d
            sg = gtot / (ytot * 1.0)
            sg.shape = (sg.size, 1)
        else:
            sg = numpy.dot(gtot, numpy.diag(1.0 / ytot))
        ng = numpy.array([sum(partition == gid) for gid in groups])
        ng.shape = (ng.size,)  # ensure ng is 1-d
        n = y.shape[0]
        # between group inequality
        sg = sg + (sg == 0)  # handle case when a partition has 0 for sum
        bg = numpy.multiply(sg, numpy.log(numpy.dot(numpy.diag(n * 1.0 / ng), sg))).sum(
            axis=0
        )

        self.T = T
        self.bg = bg
        self.wg = T - bg


class TheilDSim:
    """Random permutation based inference on Theil's inequality decomposition.
    Provides for computationally based inference regarding the inequality
    decomposition using random spatial permutations.
    See :cite:`rey_interregional_2010`.

    Parameters
    ----------

    y : numpy.array
        An array in the shape :math:`(n,t)` or :math:`(n,)`
        with :math:`n` taken as the observations across which inequality is
        calculated.  If ``y`` is :math:`(n,)` then a scalar inequality value is
        determined. If ``y`` is :math:`(n,t)` then an array of inequality values are
        determined, one value for each column in ``y``.
    partition : numpy.array
        An array in the shape :math:`(n,)` of elements indicating which partition
        each observation belongs to. These are assumed to be exhaustive.
    permutations : int
        The number of random spatial permutations for computationally
        based inference on the decomposition.

    Attributes
    ----------

    observed : numpy.array
        An array in the shape :math:`(n,t)` or :math:`(n,)`
        representing a ``TheilD`` instance for the observed data.
    bg : numpy.array
        An array in the shape ``(permutations+1, t)``
        representing between group inequality.
    bg_pvalue : numpy.array
        An array in the shape :math:`(t,1)` representing the :math:`p`-value
        for the between group measure. Measures the percentage of the realized
        values that were greater than or equal to the observed ``bg`` value.
        Includes the observed value.
    wg : numpy.array
        An array in the shape ``(permutations+1)``
        representing within group inequality. Depending on the
        shape of ``y``, the array may be 1- or 2-dimensional.

    Examples
    --------

    >>> import libpysal
    >>> import numpy
    >>> from inequality.theil import TheilDSim

    >>> f = libpysal.io.open(libpysal.examples.get_path('mexico.csv'))
    >>> vnames = [f'pcgdp{dec}' for dec in range(1940, 2010, 10)]
    >>> y = numpy.array([f.by_col[v] for v in vnames]).T
    >>> regimes = numpy.array(f.by_col('hanson98'))
    >>> numpy.random.seed(10)
    >>> theil_ds = TheilDSim(y, regimes, 999)

    >>> theil_ds.bg_pvalue
    array([0.4  , 0.344, 0.001, 0.001, 0.034, 0.072, 0.032])

    """

    def __init__(self, y, partition, permutations=99):
        observed = TheilD(y, partition)
        bg_ct = observed.bg == observed.bg  # already have one extreme value
        bg_ct = bg_ct * 1.0
        results = [observed]
        for _ in range(permutations):
            yp = numpy.random.permutation(y)
            t = TheilD(yp, partition)
            bg_ct += 1.0 * t.bg >= observed.bg
            results.append(t)
        self.results = results
        self.T = observed.T
        self.bg_pvalue = bg_ct / (permutations * 1.0 + 1)
        self.bg = numpy.array([r.bg for r in results])
        self.wg = numpy.array([r.wg for r in results])



"""
:mod:`inequality` --- Spatial Inequality Analysis
=================================================

"""

import contextlib
from importlib.metadata import PackageNotFoundError, version

from . import gini, theil
from ._indices import (
    abundance,
    ellison_glaeser_egg,
    ellison_glaeser_egg_pop,
    fractionalization_gs,
    gini_gi,
    gini_gi_m,
    gini_gig,
    herfindahl_hd,
    hoover_hi,
    isolation_ii,
    isolation_isg,
    margalev_md,
    maurel_sedillot_msg,
    maurel_sedillot_msg_pop,
    menhinick_mi,
    modified_segregation_msg,
    polarization,
    segregation_gsg,
    shannon_se,
    similarity_w_wd,
    simpson_sd,
    simpson_so,
    theil_th,
)

with contextlib.suppress(PackageNotFoundError):
    __version__ = version("inequality")



"""
Diversity indices as suggested in Nijkamp & Poot (2013) [1] and Nijkamp & Poot
(2015) [2]

References
----------

[1]_ Nijkamp & Poot
[2]_ Nijkamp, P. and Poot, J. "Cultural Diversity: A Matter of Measurement".
     IZA Discussion Paper Series No. 8782
     :cite:`nijkamp2015cultural`
     https://www.econstor.eu/bitstream/10419/107568/1/dp8782.pdf
"""

import itertools

import numpy

SMALL = numpy.finfo("float").tiny


def abundance(x):
    """
    Abundance index. :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    a : float
        Abundance index.

    Examples
    --------

    >>> import numpy
    >>> x = numpy.array([[0, 1, 2], [0, 2, 4], [0, 0, 3]])
    >>> abundance(x)
    2

    """

    xs = x.sum(axis=0)
    a = numpy.sum([1 for i in xs if i > 0])
    return a


def margalev_md(x):
    """
    Margalev MD index. :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    mmd : float
        Margalev MD index.

    Examples
    --------

    >>> import numpy
    >>> x = numpy.array([[0, 1, 2], [0, 2, 4], [0, 0, 3]])
    >>> margalev_md(x)
    0.40242960438184466

    """

    a = abundance(x)
    mmd = (a - 1.0) / numpy.log(x.sum())
    return mmd


def menhinick_mi(x):
    """
    Menhinick MI index. :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    mmi : float
        Menhinick MI index.

    Examples
    --------

    >>> import numpy
    >>> x = numpy.array([[0, 1, 2], [0, 2, 4], [0, 0, 3]])
    >>> menhinick_mi(x)
    0.2886751345948129

    """

    a = abundance(x)
    mmi = (a - 1.0) / numpy.sqrt(x.sum())
    return mmi


def simpson_so(x):
    """
    Simpson diversity index SO. :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    sso : float
        Simpson diversity index SO.

    Examples
    --------

    >>> import numpy
    >>> x = numpy.array([[0, 1, 2], [0, 2, 4], [0, 0, 3]])
    >>> simpson_so(x)
    0.5909090909090909

    """

    xs0 = x.sum(axis=0)
    xs = x.sum()
    num = (xs0 * (xs0 - 1.0)).sum()
    den = xs * (xs - 1.0)
    sso = num / den
    return sso


def simpson_sd(x):
    """
    Simpson diversity index SD. :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    ssd : float
        Simpson diversity index SD.

    Examples
    --------

    >>> import numpy
    >>> x = numpy.array([[0, 1, 2], [0, 2, 4], [0, 0, 3]])
    >>> simpson_sd(x)
    0.40909090909090906

    """

    ssd = 1.0 - simpson_so(x)
    return ssd


def herfindahl_hd(x):
    """
    Herfindahl index HD. :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    hhd : float
        Herfindahl index HD.

    Examples
    --------

    >>> import numpy
    >>> x = numpy.array([[0, 1, 2], [0, 2, 4], [0, 0, 3]])
    >>> herfindahl_hd(x)
    0.625

    """

    pgs = x.sum(axis=0)
    p = pgs.sum()
    hhd = ((pgs * 1.0 / p) ** 2).sum()
    return hhd


def theil_th(x, ridz=True):
    """
    Theil index TH as expressed in equation (32) of [2]. :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).
    ridz : bool (default True)
        Flag to add a small amount to zero values to avoid zero division problems.

    Returns
    -------

    tth : float
        Theil index TH.

    Examples
    --------

    >>> import numpy
    >>> x = numpy.array([[0, 1, 2], [0, 2, 4], [0, 0, 3]])
    >>> theil_th(x)
    0.15106563978903298

    """

    if ridz:
        x = x + SMALL * (x == 0)  # can't have 0 values
    pa = x.sum(axis=1).astype(float)  # Area totals
    pg = x.sum(axis=0).astype(float)  # Group totals
    p = pa.sum()
    num = (x / pa[:, None]) * (numpy.log(pg / p) - numpy.log(x / pa[:, None]))
    den = ((pg / p) * numpy.log(pg / p)).sum()
    th = (pa / p)[:, None] * (num / den)
    tth = th.sum().sum()
    return tth


def theil_th_brute(x, ridz=True):
    """
    Theil index TH using inefficient computation.
    NOTE: just for result comparison, it matches ``theil_th``.

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).
    ridz : bool (default True)
        Flag to add a small amount to zero values to avoid zero division problems.

    Returns
    -------

    tth : float
        Theil index TH.

    """

    if ridz:
        x = x + SMALL * (x == 0)  # can't have 0 values
    pas = x.sum(axis=1).astype(float)  # Area totals
    pgs = x.sum(axis=0).astype(float)  # Group totals
    p = pas.sum()
    th = numpy.zeros(x.shape)
    for g in numpy.arange(x.shape[1]):
        pg = pgs[g]
        for a in numpy.arange(x.shape[0]):
            pa = pas[a]
            pga = x[a, g]
            num = (pga / pa) * ((numpy.log(pg / p)) - numpy.log(pga / pa))
            den = ((pgs / p) * numpy.log(pgs / p)).sum()
            th[a, g] = (pa / p) * (num / den)
    tth = th.sum().sum()
    return tth


def fractionalization_gs(x):
    """
    Fractionalization Gini-Simpson index GS. :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    fgs : float
        Fractionalization Gini-Simpson index GS.

    Examples
    --------

    >>> import numpy
    >>> x = numpy.array([[0, 1, 2], [0, 2, 4], [0, 0, 3]])
    >>> fractionalization_gs(x)
    0.375

    """

    fgs = 1.0 - herfindahl_hd(x)
    return fgs


def polarization(x):  # noqa ARG001
    raise RuntimeError("Not currently implemented.")


def shannon_se(x):
    """
    Shannon index SE. :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    sse : float
        Shannon index SE.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> shannon_se(y)
    1.094070862104929

    """

    pgs = x.sum(axis=0)
    p = pgs.sum()
    ratios = pgs * 1.0 / p
    sse = -(ratios * numpy.log(ratios)).sum()
    return sse


def _gini(ys):
    """Gini for a single row to be used both by ``gini_gi`` and ``gini_gig``."""

    n = ys.flatten().shape[0]
    ys.sort()
    num = 2.0 * ((numpy.arange(n) + 1) * ys).sum()
    den = n * ys.sum()
    return (num / den) - ((n + 1.0) / n)


def gini_gi(x):
    """
    Gini GI index. :cite:`nijkamp2015cultural`

    NOTE: based on 3rd eq. of "Calculation" in:

            http://en.wikipedia.org/wiki/Gini_coefficient

    Returns same value as ``gini`` method in the R package ``reldist`` (see
    http://rss.acs.unt.edu/Rdoc/library/reldist/html/gini.html) if every
    category has at least one observation.

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    ggi : float
        Gini GI index.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> round(gini_gi(y), 10)
    0.0512820513

    """
    ys = x.sum(axis=0)
    return _gini(ys)


def gini_gig(x):
    """
    Gini GI index. :cite:`nijkamp2015cultural`

    NOTE: based on Wolfram Mathworld formula in:

            http://mathworld.wolfram.com/GiniCoefficient.html

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    ggig : numpy.array
        Gini GI index for every group :math:`k`.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> gini_gig(y)
    array([0.125     , 0.32894737, 0.18181818])

    """

    ggig = numpy.apply_along_axis(_gini, 0, x.copy())
    return ggig


def gini_gi_m(x):
    """
    Gini GI index (equivalent to ``gini_gi``, not vectorized).
    :cite:`nijkamp2015cultural`

    NOTE: based on Wolfram Mathworld formula in:

            http://mathworld.wolfram.com/GiniCoefficient.html

    Returns same value as ``gini_gi``.

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    ggim : float
        Gini GI index.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> round(gini_gi_m(y), 10)
    0.0512820513

    """

    xs = x.sum(axis=0)
    num = numpy.sum([numpy.abs(xi - xj) for xi, xj in itertools.permutations(xs, 2)])
    den = 2.0 * xs.shape[0] ** 2 * numpy.mean(xs)
    ggim = num / den
    return ggim


def hoover_hi(x):
    """
    Hoover index HI. :cite:`nijkamp2015cultural`

    NOTE: based on

            http://en.wikipedia.org/wiki/Hoover_index

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    hhi : float
        Hoover HI index.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> round(hoover_hi(y), 10)
    0.041025641

    """

    es = x.sum(axis=0)
    e_total = es.sum()
    a_total = es.shape[0]
    s = numpy.abs((es * 1.0 / e_total) - (1.0 / a_total)).sum()
    hhi = s / 2.0
    return hhi


def similarity_w_wd(x, tau):
    """
    Similarity weighted diversity. :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).
    tau : numpy.array
        A :math:`(k, k)` array where :math:`tau_{ij}` represents dissimilarity
        between group :math:`i` and group :math:`j`. Diagonal elements are
        assumed to be one.

    Returns
    -------

    swwd : float
        Similarity weighted diversity index.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> numpy.random.seed(0)
    >>> tau = numpy.random.uniform(size=(3,3))
    >>> numpy.fill_diagonal(tau, 0.)
    >>> tau = (tau + tau.T)/2
    >>> tau
    array([[0.        , 0.63003627, 0.52017529],
           [0.63003627, 0.        , 0.76883356],
           [0.52017529, 0.76883356, 0.        ]])

    >>> round(similarity_w_wd(y, tau), 10)
    0.581859634

    """

    pgs = x.sum(axis=0)
    pgs = pgs * 1.0 / pgs.sum()
    s = sum(
        [
            pgs[i] * pgs[j] * tau[i, j]
            for i, j in itertools.product(numpy.arange(pgs.shape[0]), repeat=2)
        ]
    )
    swwd = 1.0 - s
    return swwd


def segregation_gsg(x):
    """
    Segregation index GS.

    This is a Duncan&Duncan index of a group against the rest combined.

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    sgsg : array
        An array with GSg indices for the :math:`k` groups.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> segregation_gsg(y).round(6)
    array([0.182927, 0.24714 , 0.097252])

    """

    pgs = x.sum(axis=0)
    pas = x.sum(axis=1)
    p = pgs.sum()
    first = (x.T * 1.0 / pgs[:, None]).T
    pampga = pas[:, None] - x
    pmpg = p - pgs
    second = pampga * 1.0 / pmpg[None, :]
    sgsg = 0.5 * (numpy.abs(first - second)).sum(axis=0)
    return sgsg


def modified_segregation_msg(x):
    """
    Modified segregation index GS.

    This is a modified version of GSg index as used by Van Mourik et al. (1989)
    :cite:`van_Mourik_1989`.

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    ms_inds : numpy.array
        An array with MSg indices for the :math:`k` groups.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> modified_segregation_msg(y).round(6)
    array([0.085207, 0.102249, 0.04355 ])

    """

    pgs = x.sum(axis=0)
    p = pgs.sum()
    ms_inds = segregation_gsg(x)  # To be updated in loop below
    for gi in numpy.arange(x.shape[1]):
        pg = pgs[gi]
        pgp = pg * 1.0 / p
        ms_inds[gi] = 2.0 * pgp * (1.0 - pgp) * ms_inds[gi]
    return ms_inds


def isolation_isg(x):
    """
    Isolation index IS. :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    iisg : numpy.array
        An array with ISg indices for the :math:`k` groups.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> isolation_isg(y).round(6)
    array([1.07327 , 1.219953, 1.022711])

    """

    ws = x * 1.0 / x.sum(axis=0)
    pgapa = (x.T * 1.0 / x.sum(axis=1)).T
    pgp = x.sum(axis=0) * 1.0 / x.sum()
    iisg = (ws * pgapa / pgp).sum(axis=0)
    return iisg


def isolation_ii(x):
    """
    Isolation index :math:`II_g` as in equation (23) of [2].
    :cite:`nijkamp2015cultural`

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    iso_ii : numpy.array
        An array with IIg indices for the :math:`k` groups.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> isolation_ii(y).round(6)
    array([1.11616 , 1.310804, 1.03433 ])

    """

    pa = x.sum(axis=1).astype(float)  # Area totals
    pg = x.sum(axis=0).astype(float)  # Group totals
    p = pa.sum()
    ws = x / pg

    block = (ws * (x / pa[:, None])).sum(axis=0)
    num = (block / (pg / p)) - (pg / p)
    den = 1.0 - (pg / p)
    iso_ii = num / den
    return iso_ii


def ellison_glaeser_egg(x, hs=None):
    """
    Ellison and Glaeser (1997) :cite:`ellison_1997` index of concentration.
    Implemented as in equation (5) of original reference.

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        area) and :math:`k` columns (one per industry). Each cell indicates
        employment figures for area :math:`n` and industry :math:`k`.
    hs : numpy.array (default None)
        An array of dimension :math:`(k,)` containing the Herfindahl
        indices of each industry's plant sizes. If not passed, it is
        assumed every plant contains one and only one worker and thus
        :math:`H_k = 1 / P_k`, where :math:`P_k` is the total
        employment in :math:`k`.

    Returns
    -------

    eg_inds : numpy.array
        EG index for each of the :math:`k` groups.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> z = numpy.random.randint(10, 50, size=(3,4))
    >>> z
    array([[10, 13, 13, 49],
           [19, 29, 31, 46],
           [33, 16, 34, 34]])

    >>> ellison_glaeser_egg(z).round(6)
    array([0.054499, 0.016242, 0.010141, 0.028803])

    >>> numpy.random.seed(0)
    >>> v = numpy.random.uniform(0, 1, size=(4,)).round(3)
    >>> v
    array([0.549, 0.715, 0.603, 0.545])

    >>> ellison_glaeser_egg(z, hs=v).round(6)
    array([-1.06264 , -2.39227 , -1.461383, -1.117953])

    References
    ----------

    - :cite:`ellison_1997` Ellison, G. and Glaeser, E. L. "Geographic Concentration in U.S. Manufacturing Industries: A Dartboard Approach". Journal of Political Economy. 105: 889-927

    """  # noqa E501

    industry_totals = x.sum(axis=0)
    if hs is None:
        hs = 1.0 / industry_totals
    xs = x.sum(axis=1) * 1.0 / x.sum()
    part = 1.0 - (xs**2).sum()
    eg_inds = numpy.zeros(x.shape[1])
    for gi in numpy.arange(x.shape[1]):
        ss = x[:, gi] * 1.0 / industry_totals[gi]
        g = ((ss - xs) ** 2).sum()
        h = hs[gi]
        eg_inds[gi] = (g - part * h) / (part * (1.0 - h))
    return eg_inds


def ellison_glaeser_egg_pop(x):
    """
    Ellison and Glaeser (1997) :cite:`ellison_1997` index of concentration.
    Implemented to be computed with data about people (segregation/diversity)
    rather than as industry concentration, following Mare et al (2012)
    :cite:`care_2012`.

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    eg_inds : numpy.array
        EG index for each of the :math:`k` groups.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> ellison_glaeser_egg_pop(y).round(6)
    array([-0.021508,  0.013299, -0.038946])

    References
    ----------

    - :cite:`ellison_1997` – Ellison, G. and Glaeser, E. L. "Geographic Concentration in U.S. Manufacturing Industries: A Dartboard Approach". Journal of Political Economy. 105: 889-927

    - :cite:`care_2012` – Care, D., Pinkerton, R., Poot, J. and Coleman, A. (2012) "Residential sorting across Auckland neighbourhoods." New Zealand Population Review, 38, 23-54.

    """  # noqa E501

    pas = x.sum(axis=1)
    pgs = x.sum(axis=0)
    p = pas.sum()
    pap = pas * 1.0 / p
    opg = 1.0 / pgs
    oopg = 1.0 - opg
    eg_inds = numpy.zeros(x.shape[1])
    for g in numpy.arange(x.shape[1]):
        pgas = x[:, g]
        pg = pgs[g]
        num1n = (((pgas * 1.0 / pg) - pap) ** 2).sum()
        num1d = 1.0 - (pap**2).sum()
        num2 = opg[g]
        den = oopg[g]
        eg_inds[g] = ((num1n / num1d) - num2) / den
    return eg_inds


def maurel_sedillot_msg(x, hs=None):
    """
    Maurel and Sedillot (1999) :cite:`maurel_1999` index of concentration.
    Implemented as in equation (7) of original reference.

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        area) and :math:`k` columns (one per industry). Each cell indicates
        employment figures for area :math:`n` and industry :math:`k`.
    hs : numpy.array (default None)
        An array of dimension :math:`(k,)` containing the Herfindahl
        indices of each industry's plant sizes. If not passed, it is
        assumed every plant contains one and only one worker and thus
        :math:`H_k = 1 / P_k`, where :math:`P_k` is the total
        employment in :math:`k`.

    Returns
    -------

    ms_inds : numpy.array
        MS index for each of the :math:`k` groups.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> z = numpy.random.randint(10, 50, size=(3,4))
    >>> z
    array([[10, 13, 13, 49],
           [19, 29, 31, 46],
           [33, 16, 34, 34]])

    >>> maurel_sedillot_msg(z).round(6)
    array([ 0.078583,  0.035977,  0.039374, -0.009049])

    >>> numpy.random.seed(0)
    >>> v = numpy.random.uniform(0, 1, size=(4,)).round(3)
    >>> v
    array([0.549, 0.715, 0.603, 0.545])

    >>> maurel_sedillot_msg(z, hs=v).round(6)
    array([-1.010102, -2.324216, -1.38869 , -1.200499])

    References
    ----------

    - :cite:`maurel_1999` – Maurel, F. and Sédillot, B. (1999). "A Measure of the Geographic Concentration in French Manufacturing Industries." Regional Science and Urban Economics 29: 575-604.

    """  # noqa E501

    industry_totals = x.sum(axis=0)
    if hs is None:
        hs = 1.0 / industry_totals
    x2s = numpy.sum((x.sum(axis=1) * 1.0 / x.sum()) ** 2)
    ms_inds = numpy.zeros(x.shape[1])
    for gi in numpy.arange(x.shape[1]):
        s2s = numpy.sum((x[:, gi] * 1.0 / industry_totals[gi]) ** 2)
        h = hs[gi]
        num = ((s2s - x2s) / (1.0 - x2s)) - h
        den = 1.0 - h
        ms_inds[gi] = num / den
    return ms_inds


def maurel_sedillot_msg_pop(x):
    """
    Maurel and Sedillot (1999) :cite:`maurel_1999` index of concentration.
    Implemented to be computed with data about people (segregation/diversity)
    rather than as industry concentration, following Mare et al (2012)
    :cite:`care_2012`.

    Parameters
    ----------

    x : numpy.array
        An :math:`(N, k)` shaped array containing :math:`N` rows (one per
        neighborhood) and :math:`k` columns (one per cultural group).

    Returns
    -------

    eg_inds : numpy.array
        MS index for each of the :math:`k` groups.

    Examples
    --------

    >>> import numpy
    >>> numpy.random.seed(0)
    >>> y = numpy.random.randint(1, 10, size=(4,3))
    >>> y
    array([[6, 1, 4],
           [4, 8, 4],
           [6, 3, 5],
           [8, 7, 9]])

    >>> maurel_sedillot_msg_pop(y).round(6)
    array([-0.055036,  0.044147, -0.028666])

    References
    ----------

    - :cite:`maurel_1999` – Maurel, F. and Sédillot, B. (1999). "A Measure of the Geographic Concentration in French Manufacturing Industries." Regional Science and Urban Economics 29: 575-604.

    - :cite:`care_2012` – Care, D., Pinkerton, R., Poot, J. and Coleman, A. (2012) "Residential sorting across Auckland neighbourhoods." New Zealand Population Review, 38, 23-54.

    """  # noqa E501

    pas = x.sum(axis=1)
    pgs = x.sum(axis=0)
    p = pas.sum()
    pap = pas * 1.0 / p
    eg_inds = numpy.zeros(x.shape[1])
    for g in numpy.arange(x.shape[1]):
        pgas = x[:, g]
        pg = pgs[g]
        num1n = ((pgas * 1.0 / pg) ** 2 - pap**2).sum()
        num1d = 1.0 - (pap**2).sum()
        num2 = 1.0 / pg
        den = 1.0 - (1.0 / pg)
        eg_inds[g] = ((num1n / num1d) - num2) / den
    return eg_inds



import libpysal
import numpy

from inequality.gini import Gini, Gini_Spatial


class TestGini:
    def setup_method(self):
        f = libpysal.io.open(libpysal.examples.get_path("mexico.csv"))
        vnames = [f"pcgdp{dec}" for dec in range(1940, 2010, 10)]
        y = numpy.transpose(numpy.array([f.by_col[v] for v in vnames]))
        self.y = y[:, 0]
        regimes = numpy.array(f.by_col("hanson98"))

        self.w = libpysal.weights.block_weights(regimes, silence_warnings=True)

    def test_Gini(self):
        g = Gini(self.y)
        numpy.testing.assert_almost_equal(g.g, 0.35372371173452849)

    def test_Gini_Spatial(self):
        numpy.random.seed(12345)
        g = Gini_Spatial(self.y, self.w)
        numpy.testing.assert_almost_equal(g.g, 0.35372371173452849)
        numpy.testing.assert_almost_equal(g.wg, 884130.0)
        numpy.testing.assert_almost_equal(g.wcg, 4353856.0)
        numpy.testing.assert_almost_equal(g.p_sim, 0.040)
        numpy.testing.assert_almost_equal(g.e_wcg, 4170356.7474747472)
